@@ -1,5 +1,6 @@
 /**
  * Copyright (c) 2010 LxDE Developers, see the file AUTHORS for details.
+ * Copyright (C) 2020 Ingo Brückl
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -19,10 +20,12 @@
 #include <config.h>
 #include <locale.h>
 #include <stdlib.h>
+#include <gdk/gdkkeysyms.h>
 #include <gtk/gtk.h>
 #include <gdk/gdk.h>
 #include <gdk/gdkx.h>
 #include <glib/gi18n.h>
+#include <sys/file.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <sys/wait.h>
@@ -36,18 +39,20 @@
 #include <X11/Xatom.h>
 #include <X11/Xlib.h>
 
-#include "dbus-interface.h"
+#include "lxsession-logout-dbus-interface.h"
 
 /* Command parameters. */
 static char * prompt = NULL;
 static char * banner_side = NULL;
 static char * banner_path = NULL;
+static char **hide_button = NULL;
 
 static GOptionEntry opt_entries[] =
 {
     { "prompt", 'p', 0, G_OPTION_ARG_STRING, &prompt, N_("Custom message to show on the dialog"), N_("message") },
     { "banner", 'b', 0, G_OPTION_ARG_STRING, &banner_path, N_("Banner to show on the dialog"), N_("image file") },
     { "side", 's', 0, G_OPTION_ARG_STRING, &banner_side, N_("Position of the banner"), "top|left|right|bottom" },
+    { "hide-button", 'H', 0, G_OPTION_ARG_STRING_ARRAY, &hide_button, N_("Hide button"), "shutdown|reboot|logout|hibernate|suspend|switch_user|lock_screen" },
     { NULL }
 };
 
@@ -61,18 +66,16 @@ typedef struct {
     int hibernate_available : 1;	/* Hibernate is available */
     int switch_user_available : 1;	/* Switch User is available */
 
-    int shutdown_logind : 1;		/* Shutdown is available via logind */
-    int reboot_logind : 1;		/* Reboot is available via logind */
-    int suspend_logind : 1;		/* Suspend is available via logind */
-    int hibernate_logind : 1;		/* Hibernate is available via logind */
+    int shutdown_systemd : 1;		/* Shutdown is available via systemd */
+    int reboot_systemd : 1;		/* Reboot is available via systemd */
+    int suspend_systemd : 1;		/* Suspend is available via systemd */
+    int hibernate_systemd : 1;		/* Hibernate is available via systemd */
     int shutdown_ConsoleKit : 1;	/* Shutdown is available via ConsoleKit */
     int reboot_ConsoleKit : 1;		/* Reboot is available via ConsoleKit */
+    int suspend_ConsoleKit : 1;		/* Suspend is available via ConsoleKit */
+    int hibernate_ConsoleKit : 1;	/* Hibernate is available via ConsoleKit */
     int suspend_UPower : 1;		/* Suspend is available via UPower */
     int hibernate_UPower : 1;		/* Hibernate is available via UPower */
-    int shutdown_HAL : 1;		/* Shutdown is available via HAL */
-    int reboot_HAL : 1;			/* Reboot is available via HAL */
-    int suspend_HAL : 1;		/* Suspend is available via HAL */
-    int hibernate_HAL : 1;		/* Hibernate is available via HAL */
     int switch_user_GDM : 1;		/* Switch User is available via GDM */
     int switch_user_LIGHTDM : 1;	/* Switch User is available via GDM */
     int switch_user_KDM : 1;		/* Switch User is available via LIGHTDM */
@@ -96,7 +99,11 @@ static void switch_user_clicked(GtkButton * button, HandlerContext * handler_con
 static void cancel_clicked(GtkButton * button, gpointer user_data);
 static GtkPositionType get_banner_position(void);
 static GdkPixbuf * get_background_pixbuf(void);
+#ifdef USE_GTK3
+gboolean draw(GtkWidget * widget, cairo_t * cr, GdkPixbuf * pixbuf);
+#else
 gboolean expose_event(GtkWidget * widget, GdkEventExpose * event, GdkPixbuf * pixbuf);
+#endif
 
 /* Try to run lxlock command in order to lock the screen, return TRUE on
  * success, FALSE if command execution failed
@@ -142,6 +149,9 @@ static gboolean verify_running(const char * display_manager, const char * execut
         char buffer[PATH_MAX];
         sprintf(buffer, "/var/run/%s.pid", display_manager);
 
+        if (!g_file_test (buffer, G_FILE_TEST_IS_REGULAR))
+            sprintf(buffer, "/var/run/%s/%s.pid", display_manager, display_manager);
+
         /* Open the pid file. */
         int fd = open(buffer, O_RDONLY);
         if (fd >= 0)
@@ -212,7 +222,7 @@ static void change_root_property(GtkWidget* w, const char* prop_name, const char
 /* Handler for "clicked" signal on Shutdown button. */
 static void shutdown_clicked(GtkButton * button, HandlerContext * handler_context)
 {
-    char * error_result = NULL;
+    GError *err = NULL;
     gtk_label_set_text(GTK_LABEL(handler_context->error_label), NULL);
 
     if (handler_context->ltsp)
@@ -224,21 +234,25 @@ static void shutdown_clicked(GtkButton * button, HandlerContext * handler_contex
         }
     }
     else if (handler_context->shutdown_ConsoleKit)
-        error_result = dbus_ConsoleKit_Stop();
-    else if (handler_context->shutdown_logind)
-        error_result = dbus_logind_PowerOff();
-    else if (handler_context->shutdown_HAL)
-        error_result = dbus_HAL_Shutdown();
+        dbus_ConsoleKit_PowerOff(&err);
+    else if (handler_context->shutdown_systemd)
+        dbus_systemd_PowerOff(&err);
 
-    if (error_result != NULL)
-        gtk_label_set_text(GTK_LABEL(handler_context->error_label), error_result);
-        else gtk_main_quit();
+	if (err)
+	{
+		gtk_label_set_text(GTK_LABEL(handler_context->error_label), err->message);
+		g_error_free (err);
+	}
+	else
+    {
+        gtk_main_quit();
+    }
 }
 
 /* Handler for "clicked" signal on Reboot button. */
 static void reboot_clicked(GtkButton * button, HandlerContext * handler_context)
 {
-    char * error_result = NULL;
+    GError *err = NULL;
     gtk_label_set_text(GTK_LABEL(handler_context->error_label), NULL);
 
     if (handler_context->ltsp)
@@ -250,58 +264,75 @@ static void reboot_clicked(GtkButton * button, HandlerContext * handler_context)
         }
     }
     else if (handler_context->reboot_ConsoleKit)
-        error_result = dbus_ConsoleKit_Restart();
-    else if (handler_context->reboot_logind)
-        error_result = dbus_logind_Reboot();
-    else if (handler_context->reboot_HAL)
-        error_result = dbus_HAL_Reboot();
+        dbus_ConsoleKit_Reboot(&err);
+    else if (handler_context->reboot_systemd)
+        dbus_systemd_Reboot(&err);
 
-    if (error_result != NULL)
-        gtk_label_set_text(GTK_LABEL(handler_context->error_label), error_result);
-        else gtk_main_quit();
+	if (err)
+	{
+		gtk_label_set_text(GTK_LABEL(handler_context->error_label), err->message);
+		g_error_free (err);
+	}
+	else
+    {
+        gtk_main_quit();
+    }
 }
 
 /* Handler for "clicked" signal on Suspend button. */
 static void suspend_clicked(GtkButton * button, HandlerContext * handler_context)
 {
-    char * error_result = NULL;
+    GError *err = NULL;
     gtk_label_set_text(GTK_LABEL(handler_context->error_label), NULL);
 
     lock_screen();
     if (handler_context->suspend_UPower)
-        error_result = dbus_UPower_Suspend();
-    else if (handler_context->suspend_logind)
-        error_result = dbus_logind_Suspend();
-    else if (handler_context->suspend_HAL)
-        error_result = dbus_HAL_Suspend();
+        dbus_UPower_Suspend(&err);
+    else if (handler_context->suspend_ConsoleKit)
+        dbus_ConsoleKit_Suspend(&err);
+    else if (handler_context->suspend_systemd)
+        dbus_systemd_Suspend(&err);
 
-    if (error_result != NULL)
-        gtk_label_set_text(GTK_LABEL(handler_context->error_label), error_result);
-        else gtk_main_quit();
+	if (err)
+	{
+		gtk_label_set_text(GTK_LABEL(handler_context->error_label), err->message);
+		g_error_free (err);
+	}
+	else
+    {
+        gtk_main_quit();
+    }
 }
 
 /* Handler for "clicked" signal on Hibernate button. */
 static void hibernate_clicked(GtkButton * button, HandlerContext * handler_context)
 {
-    char * error_result = NULL;
+    GError *err = NULL;
     gtk_label_set_text(GTK_LABEL(handler_context->error_label), NULL);
 
     lock_screen();
     if (handler_context->hibernate_UPower)
-        error_result = dbus_UPower_Hibernate();
-    else if (handler_context->hibernate_logind)
-        error_result = dbus_logind_Hibernate();
-    else if (handler_context->hibernate_HAL)
-        error_result = dbus_HAL_Hibernate();
+        dbus_UPower_Hibernate(&err);
+    else if (handler_context->hibernate_ConsoleKit)
+        dbus_ConsoleKit_Hibernate(&err);
+    else if (handler_context->hibernate_systemd)
+        dbus_systemd_Hibernate(&err);
 
-    if (error_result != NULL)
-        gtk_label_set_text(GTK_LABEL(handler_context->error_label), error_result);
-        else gtk_main_quit();
+	if (err)
+	{
+		gtk_label_set_text(GTK_LABEL(handler_context->error_label), err->message);
+		g_error_free (err);
+	}
+	else
+    {
+        gtk_main_quit();
+    }
 }
 
 /* Handler for "clicked" signal on Switch User button. */
 static void switch_user_clicked(GtkButton * button, HandlerContext * handler_context)
 {
+    GError *err = NULL;
     gtk_label_set_text(GTK_LABEL(handler_context->error_label), NULL);
 
     lock_screen();
@@ -310,11 +341,19 @@ static void switch_user_clicked(GtkButton * button, HandlerContext * handler_con
     else if (handler_context->switch_user_KDM)
         g_spawn_command_line_sync("kdmctl reserve", NULL, NULL, NULL, NULL);
     else if (handler_context->switch_user_LIGHTDM)
-        dbus_Lightdm_SwitchToGreeter();
+        dbus_Lightdm_SwitchToGreeter(&err);
     else if(handler_context->switch_user_LXDM)
         g_spawn_command_line_sync("lxdm-binary -c USER_SWITCH", NULL, NULL, NULL, NULL);
 
-    gtk_main_quit();
+	if (err)
+	{
+		gtk_label_set_text(GTK_LABEL(handler_context->error_label), err->message);
+		g_error_free (err);
+	}
+	else
+    {
+        gtk_main_quit();
+    }
 }
 
 /* Handler for "clicked" signal on Lock button. */
@@ -352,7 +391,7 @@ static GdkPixbuf * get_background_pixbuf(void)
 {
     /* Get the root window pixmap. */
     GdkScreen * screen = gdk_screen_get_default();
-#ifdef ENABLE_GTK3
+#ifdef USE_GTK3
     GdkPixbuf * pixbuf = gdk_pixbuf_get_from_window(
         gdk_get_default_root_window(),
         0,
@@ -396,32 +435,53 @@ static GdkPixbuf * get_background_pixbuf(void)
 }
 
 /* Handler for "expose_event" on background. */
+#ifdef USE_GTK3
+gboolean draw(GtkWidget * widget, cairo_t * cr, GdkPixbuf * pixbuf)
+#else
 gboolean expose_event(GtkWidget * widget, GdkEventExpose * event, GdkPixbuf * pixbuf)
+#endif
 {
+    gint x, y;
+
     if (pixbuf != NULL)
     {
         /* Copy the appropriate rectangle of the root window pixmap to the drawing area.
          * All drawing areas are immediate children of the toplevel window, so the allocation yields the source coordinates directly. */
-#if GTK_CHECK_VERSION(2,14,0)
+#ifdef USE_GTK3
+#elif GTK_CHECK_VERSION(2,14,0)
        cairo_t * cr = gdk_cairo_create (gtk_widget_get_window(widget));
+       gdk_window_get_origin(gtk_widget_get_window(widget), &x, &y);
 #else
        cairo_t * cr = gdk_cairo_create (widget->window);
+       gdk_window_get_origin(widget->window, &x, &y);
 #endif
        gdk_cairo_set_source_pixbuf (
            cr,
            pixbuf,
-           0,
-           0);
+           -x,
+           -y);
 
        cairo_paint (cr);
+#ifndef USE_GTK3
        cairo_destroy(cr);
+#endif
     }
     return FALSE;
 }
 
+static char lockfile[PATH_MAX];
+
+/* Unlink lockfile on exit. */
+static void main_at_exit(void)
+{
+    unlink(lockfile);
+}
+
 /* Main program. */
 int main(int argc, char * argv[])
 {
+    GdkColor color;
+
 #ifdef ENABLE_NLS
     setlocale(LC_ALL, "");
     bindtextdomain(GETTEXT_PACKAGE, PACKAGE_LOCALE_DIR);
@@ -429,19 +489,6 @@ int main(int argc, char * argv[])
     textdomain (GETTEXT_PACKAGE);
 #endif
 
-    /* Initialize GTK (via g_option_context_parse) and parse command line arguments. */
-    GOptionContext * context = g_option_context_new("");
-    g_option_context_add_main_entries(context, opt_entries, GETTEXT_PACKAGE);
-    g_option_context_add_group(context, gtk_get_option_group(TRUE));
-    GError * err = NULL;
-    if ( ! g_option_context_parse(context, &argc, &argv, &err))
-    {
-        g_print(_("Error: %s\n"), err->message);
-        g_error_free(err);
-        return 1;
-    }
-    g_option_context_free(context);
-
     HandlerContext handler_context;
     memset(&handler_context, 0, sizeof(handler_context));
 
@@ -449,39 +496,63 @@ int main(int argc, char * argv[])
     const char * p = g_getenv("_LXSESSION_PID");
     if (p != NULL) handler_context.lxsession_pid = atoi(p);
 
-    /* Initialize capabilities of the logind mechanism. */
-    if (dbus_logind_CanPowerOff())
+    /* Create lock file to prevent more than one logout dialog per lxsession process. */
+    sprintf(lockfile, "/tmp/.lxsession-logout-%d.lock", handler_context.lxsession_pid);
+    int fd = open(lockfile, O_RDONLY|O_CREAT, 00600);
+    if (fd >= 0)
+    {
+        if (flock(fd, LOCK_EX | LOCK_NB))
+        {
+            exit(EXIT_FAILURE);
+        }
+    }
+    atexit(main_at_exit);
+
+    /* Query DBus before GTK+ initialization!!! Otherwise a race may occur. */
+
+    /* Initialize capabilities of the systemd mechanism. */
+    if (dbus_systemd_CanPowerOff())
     {
         handler_context.shutdown_available = TRUE;
-        handler_context.shutdown_logind = TRUE;
+        handler_context.shutdown_systemd = TRUE;
     }
-    if (dbus_logind_CanReboot())
+    if (dbus_systemd_CanReboot())
     {
         handler_context.reboot_available = TRUE;
-        handler_context.reboot_logind = TRUE;
+        handler_context.reboot_systemd = TRUE;
     }
-    if (dbus_logind_CanSuspend())
+    if (dbus_systemd_CanSuspend())
     {
         handler_context.suspend_available = TRUE;
-        handler_context.suspend_logind = TRUE;
+        handler_context.suspend_systemd = TRUE;
     }
-    if (dbus_logind_CanHibernate())
+    if (dbus_systemd_CanHibernate())
     {
         handler_context.hibernate_available = TRUE;
-        handler_context.hibernate_logind = TRUE;
+        handler_context.hibernate_systemd = TRUE;
     }
 
     /* Initialize capabilities of the ConsoleKit mechanism. */
-    if (!handler_context.shutdown_available && dbus_ConsoleKit_CanStop())
+    if (!handler_context.shutdown_available && dbus_ConsoleKit_CanPowerOff())
     {
         handler_context.shutdown_available = TRUE;
         handler_context.shutdown_ConsoleKit = TRUE;
     }
-    if (!handler_context.reboot_available && dbus_ConsoleKit_CanRestart())
+    if (!handler_context.reboot_available && dbus_ConsoleKit_CanReboot())
     {
         handler_context.reboot_available = TRUE;
         handler_context.reboot_ConsoleKit = TRUE;
     }
+    if (!handler_context.suspend_available && dbus_ConsoleKit_CanSuspend())
+    {
+        handler_context.suspend_available = TRUE;
+        handler_context.suspend_ConsoleKit = TRUE;
+    }
+    if (!handler_context.hibernate_available && dbus_ConsoleKit_CanHibernate())
+    {
+        handler_context.hibernate_available = TRUE;
+        handler_context.hibernate_ConsoleKit = TRUE;
+    }
 
     /* Initialize capabilities of the UPower mechanism. */
     if (!handler_context.suspend_available && dbus_UPower_CanSuspend())
@@ -495,28 +566,6 @@ int main(int argc, char * argv[])
         handler_context.hibernate_UPower = TRUE;
     }
 
-    /* Initialize capabilities of the HAL mechanism. */
-    if (!handler_context.shutdown_available && dbus_HAL_CanShutdown())
-    {
-        handler_context.shutdown_available = TRUE;
-        handler_context.shutdown_HAL = TRUE;
-    }
-    if (!handler_context.reboot_available && dbus_HAL_CanReboot())
-    {
-        handler_context.reboot_available = TRUE;
-        handler_context.reboot_HAL = TRUE;
-    }
-    if (!handler_context.suspend_available && dbus_HAL_CanSuspend())
-    {
-        handler_context.suspend_available = TRUE;
-        handler_context.suspend_HAL = TRUE;
-    }
-    if (!handler_context.hibernate_available && dbus_HAL_CanHibernate())
-    {
-        handler_context.hibernate_available = TRUE;
-        handler_context.hibernate_HAL = TRUE;
-    }
-
     /* If we are under GDM, its "Switch User" is available. */
     if (verify_running("gdm", "gdmflexiserver"))
     {
@@ -531,21 +580,7 @@ int main(int argc, char * argv[])
         handler_context.switch_user_GDM = TRUE;
     }
 
-    /* lightdm also use gdmflexiserver */
-    if (verify_running("lightdm", "gdmflexiserver"))
-    {
-        handler_context.switch_user_available = TRUE;
-        handler_context.switch_user_GDM = TRUE;
-    }
-
-    /* lightdm also use gdmflexiserver */
-    if (verify_running("lightdm", "gdmflexiserver"))
-    {
-        handler_context.switch_user_available = TRUE;
-        handler_context.switch_user_GDM = TRUE;
-    }
-
-    /* lightdm can also be find by the env */
+    /* lightdm can be found by the env */
     if (g_getenv("XDG_SEAT_PATH"))
     {
         handler_context.switch_user_available = TRUE;
@@ -580,6 +615,19 @@ int main(int argc, char * argv[])
         handler_context.lock_screen = TRUE;
     }
 
+    /* Initialize GTK (via g_option_context_parse) and parse command line arguments. */
+    GOptionContext * context = g_option_context_new("");
+    g_option_context_add_main_entries(context, opt_entries, GETTEXT_PACKAGE);
+    g_option_context_add_group(context, gtk_get_option_group(TRUE));
+    GError * err = NULL;
+    if ( ! g_option_context_parse(context, &argc, &argv, &err))
+    {
+        g_print(_("Error: %s\n"), err->message);
+        g_error_free(err);
+        return 1;
+    }
+    g_option_context_free(context);
+
     /* Make the button images accessible. */
     gtk_icon_theme_append_search_path(gtk_icon_theme_get_default(), PACKAGE_DATA_DIR "/lxsession/images");
 
@@ -590,10 +638,13 @@ int main(int argc, char * argv[])
     GtkWidget * window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
     gtk_window_set_decorated(GTK_WINDOW(window), FALSE);
     gtk_window_fullscreen(GTK_WINDOW(window));
-    GdkScreen* screen = gtk_widget_get_screen(window);
-    gtk_window_set_default_size(GTK_WINDOW(window), gdk_screen_get_width(screen), gdk_screen_get_height(screen));
     gtk_widget_set_app_paintable(window, TRUE);
+#ifdef USE_GTK3
+    g_signal_connect(G_OBJECT(window), "draw", G_CALLBACK(draw), pixbuf);
+#else
     g_signal_connect(G_OBJECT(window), "expose_event", G_CALLBACK(expose_event), pixbuf);
+#endif
+    g_signal_connect(G_OBJECT(window), "destroy", G_CALLBACK(gtk_main_quit), NULL);
 
     /* Toplevel container */
     GtkWidget* alignment = gtk_alignment_new(0.5, 0.5, 0.0, 0.0);
@@ -602,12 +653,40 @@ int main(int argc, char * argv[])
     GtkWidget* center_area = gtk_event_box_new();
     gtk_container_add(GTK_CONTAINER(alignment), center_area);
 
+#ifdef USE_GTK3
+    gtk_style_context_add_class (gtk_widget_get_style_context (center_area), GTK_STYLE_CLASS_BACKGROUND);
+#endif
+
     GtkWidget* center_vbox = gtk_vbox_new(FALSE, 6);
     gtk_container_set_border_width(GTK_CONTAINER(center_vbox), 12);
     gtk_container_add(GTK_CONTAINER(center_area), center_vbox);
 
     GtkWidget* controls = gtk_vbox_new(FALSE, 6);
 
+    /* Hide buttons */
+    if (hide_button != NULL) {
+      for (guint i = 0; i < g_strv_length(hide_button); i++) {
+        if (strcmp(hide_button[i], "shutdown") == 0) {
+          handler_context.shutdown_available = FALSE;
+        }
+        if (strcmp(hide_button[i], "reboot") == 0) {
+          handler_context.reboot_available = FALSE;
+        }
+        if (strcmp(hide_button[i], "hibernate") == 0) {
+          handler_context.hibernate_available = FALSE;
+        }
+        if (strcmp(hide_button[i],"suspend") == 0) {
+          handler_context.suspend_available = FALSE;
+        }
+        if (strcmp(hide_button[i], "switch_user") == 0) {
+          handler_context.switch_user_available = FALSE;
+        }
+        if (strcmp(hide_button[i], "lock_screen") == 0) {
+          handler_context.lock_screen = FALSE;
+        }
+      }
+    }
+
     /* If specified, apply a user-specified banner image. */
     if (banner_path != NULL)
     {
@@ -664,32 +743,17 @@ int main(int argc, char * argv[])
         if (session_name == NULL)
             session_name = "LXDE";
 
-        const gchar *command_line = "lsb_release -r -s";
-        gchar *output = NULL;
-        GError *error;
+        gchar *version_id = NULL;
+#if GLIB_CHECK_VERSION(2, 64, 0)
+        version_id = g_get_os_info(G_OS_INFO_KEY_VERSION_ID);
+#endif
 
-        if (!g_spawn_command_line_sync( command_line,
-                                        &output,
-                                        NULL,
-                                        NULL,
-                                        &error))
-        {
-
-            fprintf (stderr, "Error: %s\n", error->message);
-            g_error_free (error);
-
-        }
-
-        if (output == NULL)
-        {
-            output = "";
-        }
+        if (version_id)
+            prompt = g_strdup_printf(_("<b><big>Logout %s %s session?</big></b>"), session_name, version_id);
         else
-        {
-            output[strlen ( output ) - 1] = '\0';
-        }
+            prompt = g_strdup_printf(_("<b><big>Logout %s session?</big></b>"), session_name);
 
-        prompt = g_strdup_printf(_("<b><big>Logout %s %s session ?</big></b>"), session_name, output);
+        g_free(version_id);
     }
     gtk_label_set_markup(GTK_LABEL(label), prompt);
     gtk_box_pack_start(GTK_BOX(controls), label, FALSE, FALSE, 4);
@@ -772,6 +836,15 @@ int main(int argc, char * argv[])
     GtkWidget * cancel_button = gtk_button_new_from_stock(GTK_STOCK_CANCEL);
     gtk_button_set_alignment(GTK_BUTTON(cancel_button), 0.0, 0.5);
     g_signal_connect(G_OBJECT(cancel_button), "clicked", G_CALLBACK(cancel_clicked), NULL);
+    GtkAccelGroup* accel_group = gtk_accel_group_new();
+    gtk_window_add_accel_group(GTK_WINDOW(window), accel_group);
+#if GTK_CHECK_VERSION(3,0,0)
+    gtk_widget_add_accelerator(cancel_button, "activate", accel_group,
+        GDK_KEY_Escape, (GdkModifierType)0, GTK_ACCEL_VISIBLE);
+#else
+    gtk_widget_add_accelerator(cancel_button, "activate", accel_group,
+        GDK_Escape, (GdkModifierType)NULL, GTK_ACCEL_VISIBLE);
+#endif
     gtk_box_pack_start(GTK_BOX(controls), cancel_button, FALSE, FALSE, 4);
 
     /* Create the error text. */
@@ -779,6 +852,15 @@ int main(int argc, char * argv[])
     gtk_label_set_justify(GTK_LABEL(handler_context.error_label), GTK_JUSTIFY_CENTER);
     gtk_box_pack_start(GTK_BOX(controls), handler_context.error_label, FALSE, FALSE, 4);
 
+    /* Avoid "flickering" caused by drawing the window with its normal
+     * (e.g. gray) background fullscreen at position (0,0) first and
+     * then drawing it to the center of the screen in expose_event().
+     * With a black background and window size 0 the screen is just black
+     * when entering expose_event(). */
+    gdk_color_parse("black", &color);
+    gtk_widget_modify_bg(window, GTK_STATE_NORMAL, &color);
+    gtk_widget_set_size_request(window, 0, 0);
+
     /* Show everything. */
     gtk_widget_show_all(window);
 
